#include <FractalApp.hpp>

#define GL_VER_MAJOR 2
#define GL_VER_MINOR 1

void FractalInspectorApp::OnShaderLoadFail(const std::string& type, const std::string& infoLog) {
    auto log_lines = misc::string_expand(infoLog, '\n');
    console->Log(type, Colors::Red);
    std::cerr << type << std::endl;
    for (auto line: log_lines) {
        console->Log(line, Colors::Red);
        std::cerr << line << std::endl;
    }
}

FractalInspectorApp::FractalInspectorApp(int width, int height):
    OpenGLWindow("FractalInspector",width, height, GL_VER_MAJOR, GL_VER_MINOR) {
    Shader::OnCompilationErrorMessage += [this] (auto... args) { OnShaderLoadFail(args...);};
    u_time = 0;
}

std::filesystem::path FractalInspectorApp::VertexShaderFilepathFromPrefixName(const std::string &name) {
    return "shaders/" + name + ".vert.glsl";
}

std::filesystem::path FractalInspectorApp::FragmentShaderFilepathFromPrefixName(const std::string &name) {
    return "shaders/" + name + ".frag.glsl";
}

void FractalInspectorApp::LoadShaders() {

    // TODO: Come up with a decent way to bake the shader program sources into the executable, when compiling a release build!

    mandelbrot_shader = new Shader(VertexShaderFilepathFromPrefixName("test"), FragmentShaderFilepathFromPrefixName("mandelbrot"));
    julia_shader = new Shader(VertexShaderFilepathFromPrefixName("test" ), FragmentShaderFilepathFromPrefixName("julia"));
}

void FractalInspectorApp::ReloadShader() {
    LoadShaders();
}

void FractalInspectorApp::PositionCmd(const CmdArgs &args) {
    if (args.empty()) { // Output position + scale
        console->Log(std::format("Pos: {}, {}, Zoom: {}", u_translation.x, u_translation.y, u_scale));
        return;
    }

    if (args.size() >= 2) { // Set Position
        u_translation_goal.x = std::stof(args[0]);
        u_translation_goal.y = std::stof(args[1]);
        if (args.size() >= 3) { // Also set scale
            u_scale_goal = std::stof(args[2]);
        }
    }
    console->Log(std::format("{}", args.size()));
}

void FractalInspectorApp::ZoomCmd(const CmdArgs& args) {
    if (args.empty()) {
        console->Log(std::format("Zoom: {}", u_scale)); return;
    }
}

void FractalInspectorApp::ParseCmdLineMessage(const std::string &message) {
    auto tokens = misc::string_expand(message);

    if (tokens.size() == 0) {
        console->Log("No command input!", Colors::Red);
        return;
    }

    std::string cmd = tokens[0];

    // Remove 0th element from tokens before passing to command delegates.
    tokens.erase(tokens.begin());

    if (misc::string_matches(cmd, {"r", "reload", "refresh"})) {
        return ReloadShader();
    }

    if (misc::string_matches(cmd, {"zoom"})) {
        return ZoomCmd(tokens);
    }

    if (misc::string_matches(cmd, {"pos", "position", "tp"})) {
        return PositionCmd(tokens);
    }

    if (misc::string_matches(cmd, {"reset"})) {

    }

    if (misc::string_matches(cmd, {"screenshot", "screeny", "capture"})) {
        return TakeScreenshot();
    }

    console->Log(std::format("No such command: {}", cmd), Colors::Red);
}

void FractalInspectorApp::LoadJulia() {
    juliaset_dialog->Open();
}

void FractalInspectorApp::UnloadJulia() {
    juliaset_dialog->Close();
}

// Tested and works on both platform, but is a security **NIGHTMARE**.
void OpenURL(const std::string &url) {
#ifdef _WIN32
    system(std::format("start {}", url).c_str());
#endif
#ifdef linux
    system(std::format("xdg-open {}", url).c_str());
#endif
}

JUI::Window * FractalInspectorApp::CreateAppInfoDialogWindow(JUI::Widget *parent) {
    // TODO: Implement JUI structure that makes blocks of text easy to impelement.
    auto window = new JUI::Window(parent);
    window->Title("About FractalInspector");
    window->Size({300_px, 375_px});
    window->MinSize({300, 375});

    window->Position({100_percent - 325_px, 100_percent - 400_px});

    auto* layout = new JUI::VerticalListLayout(window->Content());
    layout->Padding(0_px);

    // TODO: Code like this ends up being a common construct in JUI programs: Make a TextList widget.
    auto line_item = [&] (const std::string& text, int size = 20, const Color4& color = Colors::White) {
        auto* content = new JUI::TextRect(layout);
        content->Size({100_percent, JUI::UDim(size+5, 0)});
        content->Content(text);
        content->TextSize(size);
        content->TextColor(color);
        content->BGColor(Colors::Transparent);
        content->AlignCenterHorizontally();
        content->BorderWidth(0);
    };


    line_item("FractalInspector", 30);
    line_item("Version 1.0", 14);

    auto* box = new JUI::Rect(layout);
    box->Size({100_percent, 200_px});
    box->BGColor(Colors::Transparent);
    box->BorderWidth(0);
    auto* img = new JUI::ImageRect(box);
    img->Content(new JGL::Texture("../icon.png"));
    img->Size({200_px, 200_px});
    img->BGColor(Colors::Transparent);
    img->AnchorPoint({.5f, .5f});
    img->Position({50_percent, 50_percent});
    img->BorderWidth(0);
    img->BGColor(Colors::Transparent);

    line_item("Developed & Maintained by Josh O'Leary.", 16);
    line_item("William R. Maxine H. Ash B.", 12);
    line_item("(c) 2024 - 2025 Redacted Software", 16);

    auto* btn_box = new JUI::Rect(layout);
    btn_box->Size({100_percent, 40_px});
    btn_box->BGColor(Colors::Transparent);
    btn_box->BorderWidth(0);

    auto* btn_layout = new JUI::HorizontalListLayout(btn_box);
    btn_layout->Padding(8_px);

    auto btn_item = [&] (const std::string& text, const std::function<void()>& callback = {}) -> JUI::TextButton* {
        JUI::TextButton* btn = new JUI::TextButton(btn_layout);
        btn->Content(text);
        btn->Size({32_percent, 100_percent});
        btn->TextColor(Colors::Black);
        btn->Center();
        btn->OnClickEvent += [callback] (auto a, auto b) {
            callback();
        };
        return btn;
    };

    auto* btn_a = btn_item("License");
    auto* btn_b = btn_item("Wiki");
    auto* btn_c = btn_item("Source", []() {
        OpenURL("https://git.redacted.cc/josh/FractalInspector");
    });


    return window;
}

void FractalInspectorApp::CreateMenu() {
    using namespace JUI::UDimLiterals;

    toolbar = new JUI::UtilityBar(scene);
    auto* fractals_list = toolbar->AddSubmenu("Fractals");

    fractals_list->AddButton("Mandelbrot set", [this]{
        UnloadJulia();
        current_fractal = Fractal::MandelbrotSet;
        current_shader = mandelbrot_shader;
    });
    fractals_list->AddButton("Julia set", [this]{
        LoadJulia();
        current_fractal = Fractal::JuliaSet;
        current_shader = julia_shader;
    });



    auto* view = toolbar->AddSubmenu("View"); {
        view->AddButton("Color Picker Dialog", [this] {
            colorpicker_window->Toggle();
        });

        view->AddButton("Console", [this] {
            console->Toggle();
        });

        view->AddButton("README", [this] {
            info_dialog->Toggle();
        });

        view->AddSeparator();
        view->AddButton("Reset Viewport");
        view->AddButton("Reset Colors");
        view->AddButton("Zoom In");
        view->AddButton("Zoom Out");
    }
    auto* btn_toggle_console = toolbar->AddButton("Console");

    btn_toggle_console->OnClickEvent += [this] (auto a, auto b) mutable {
        console->Toggle();
    };

    auto* btn_toggle_info_dialog = toolbar->AddButton("Info");

    auto* misc = toolbar->AddSubmenu("Miscellaneous");
    {
        misc->AddButton("Take Screenshot", [this] () { this->TakeScreenshot(); });
        misc->AddButton("Take Collage");
    }


    info_dialog = CreateAppInfoDialogWindow(scene);

    btn_toggle_info_dialog->OnClickEvent += [this] (auto a, auto b) mutable {
        info_dialog->Toggle();
    };

    fps_label = new JUI::TextRect(toolbar);
    fps_label->Position({100_percent - 200_px, 0_px});
    fps_label->Size({200_px, 100_percent});
    fps_label->TextColor(Colors::Black);
    fps_label->Content("60 FPS");

    // TODO: Utilize for things later.
    colorpicker_window = new JUI::Window(scene);
    colorpicker_window->Title("Color-Pickers");
    colorpicker_window->MinSize({300, 150});
    colorpicker_window->Size({400_px, 150_px});
    colorpicker_window->Position({50_px, 50_px});

    auto* col_layout = new JUI::HorizontalListLayout(colorpicker_window->Content());
    //wind->Close();

    JUI::ColorPicker* pp = new JUI::ColorPicker(col_layout);
    pp->Size({25_percent, 100_percent});
    pp->OnColorValueChanged += [this] (const Color4& color) mutable {
        u_rgb_1 = color;
    };

    /*auto* wind2 = new JUI::Window(scene);
        wind2->SetTitle("Color-Picker B");
        wind2->MinSize({200, 150});
        wind2->Size({200_px, 150_px});
        wind2->Position({50_px, 250_px});*/

    JUI::ColorPicker* pp2 = new JUI::ColorPicker(col_layout);
    pp2->Size({25_percent, 100_percent});
    pp2->OnColorValueChanged += [this] (const Color4& color) mutable {
        u_rgb_2 = color;
    };

    JUI::ColorPicker* pp3 = new JUI::ColorPicker(col_layout);
    pp3->Size({25_percent, 100_percent});
    pp3->OnColorValueChanged += [this] (const Color4& color) mutable {
        u_rgb_3 = color;
    };

    JUI::ColorPicker* pp4 = new JUI::ColorPicker(col_layout);
    pp4->Size({25_percent, 100_percent});
    pp4->OnColorValueChanged += [this] (const Color4& color) mutable {
        u_rgb_3 = color;
    };


    console = new JUI::CommandLine(scene);
    console->MinSize(Vector2(200, 200));
    console->MaxSize(Vector2(GetWidth()-10, GetHeight()-10));
    console->Size({100_percent, 200_px});
    console->Position({0_percent, JUI::UDim(-200, 1)});
    console->OnInput += [this] (const std::string& message) {
        ParseCmdLineMessage(message);
    };
    console->Close();


    mandelbrotset_dialog = new JUI::Window(scene);
    mandelbrotset_dialog->Position({20_px, 300_px});
    mandelbrotset_dialog->Close();

    auto* mandelbrotset_layout = new JUI::VerticalListLayout(mandelbrotset_dialog->ViewportInstance());
    mandelbrotset_layout->LayoutOrder(JUI::LayoutOrder::V::TOP);

    int idx = 999;

    auto mb_line_item = [&] (const std::string& text, int size = 20, const Color4& color = Colors::White) {

        auto* content = new JUI::TextRect(mandelbrotset_layout);
        content->Size({100_percent, 25_px});
        content->Content(text);
        content->TextSize(size);
        content->TextColor(color);
        content->BGColor(Colors::Transparent);
        content->AlignCenterHorizontally();
        content->LayoutOrder(idx--);
    };

    mb_line_item("The Mandelbrot Set");
    mb_line_item("Imagine a simple rule: ");
    mb_line_item("Take a number, square it, and add back the original number.");

    // TODO: Collapsible section: Evaluation of mandelbrot set with z = 0.
    mb_line_item("If we start with zero, and keep repeating this rule, we always get zero. Boring, right?");

    // TODO: Collapsible section: Evaluation of mandelbrot set with z = 1.
    mb_line_item("But what if we start with a different number. Let's say 1.");
    mb_line_item("1 squared is 1, plus 1 is 2.");
    mb_line_item("2 squared is 4, plus 1 is 5.");
    mb_line_item("5 squared is 25, plus 1 is 26.");
    mb_line_item("The numbers keep getting bigger and bigger -- they 'escape' to infinity.");

    // TODO: Collapsible section: Evaluation of mandelbrot set with z = -1.
    mb_line_item("Now, let's try a number like -1.");
    mb_line_item("-1 squared is 1, plus -1 is 0.");
    mb_line_item("0 squared is 0, plus -1 is -1");
    mb_line_item("-1 squared is 1, plus -1 is 0");
    mb_line_item("The numbers get stuck in a loop between -1 and 0 -- they don't escape");

    mb_line_item("The Mandelbrot set is a collection of all the starting numbers that don't escape to infinity when you repeat this simple rule.");
    mb_line_item("These numbers, when plotted on a complex number plane, create this incredibly intricate and beautiful fractal shape.");
    mb_line_item("Color-code the 'escape velocity' of the input numbers that do escape, and you get a result like the one you see.");


    juliaset_dialog = new JUI::Window(scene);
    juliaset_dialog->Title("Julia Set Inputs");
    juliaset_dialog->Position({20_px, 300_px});
    juliaset_dialog->Size({900_px, 70_px});
    juliaset_dialog->Close(); // Defaults to close.

    auto* julia_x_slider = new JUI::Slider(juliaset_dialog->Content());

    julia_x_slider->Size({900_px, 25_px});
    julia_x_slider->Interval(1e-5f); julia_x_slider->Minimum(0); julia_x_slider->Maximum(1);
    julia_x_slider->ValueChanged += [this] (double value) {
        value = value * 2;
        value = value - 1;
        u_julia_set.x = value;
    };

    auto* julia_y_slider = new JUI::Slider(juliaset_dialog->Content());
    julia_y_slider->Size({900_px, 25_px});
    julia_y_slider->Position({0_px, 30_px});
    julia_y_slider->Interval(1e-5f); julia_y_slider->Minimum(0); julia_y_slider->Maximum(1);
    julia_y_slider->ValueChanged += [this] (double value) {
        value = value * 2;
        value = value - 1;
        u_julia_set.y = value;
    };
}

bool FractalInspectorApp::Open() {
    if (!OpenGLWindow::Open())
        return false;

    auto size = GetSize();
    auto vec_size = Vector2i(size.x, size.y);
    if (!JGL::Init(vec_size, 0.f, 1.f))
        return false;
    JGL::Update(vec_size);


    glClearColor(0.f, 0.f, 0.f, 0.f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glDepthMask(GL_TRUE);
    // TODO: Delete when we update to the next release of JGL
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1); // NOTE: This MUST be called for text rendering to work properly!!!

    scene = new JUI::Scene();
    CreateMenu();


    SampleRate canvas_samples = SampleRate::X16;
    FilteringMode canvas_filtering = FilteringMode::MIPMAP_NEAREST;

    canvas = new RenderTarget(vec_size, Colors::Transparent, false, canvas_samples, canvas_filtering);
    LoadShaders();

    console->Log(std::format("OpenGL Renderer: {}", GetGLRenderer()));
    console->Log(std::format("OpenGL Vendor: {}", GetGLVendor()));
    console->Log(std::format("OpenGL Version: {}", GetGLVersion()));
    console->Log(std::format("GLSL Version: {}", GetGLSLVersion()));
    /*auto ext_list = GetGLExtensionList();

        // TODO: output 3 at a time.
        console->Log("GL-Extensions: ");
        std::string formatted_output = "";//"GL-Extensions: {}";


        int iter = 0;
        for (auto& str : ext_list) {
            formatted_output += str + ", ";
            iter++;
            if (iter > 5) {
                iter = 0;
                console->Log(formatted_output);
                formatted_output = "";
            }
        }*/

    return true;
}

void FractalInspectorApp::PropagateWindowSize() {
    auto size = GetSize();
    Vector2i vSize = Vector2i(size.x, size.y);

    if (vSize != canvas->GetDimensions())
        canvas->Resize(vSize);

    JGL::Update(vSize);
    scene->SetViewportSize(Vector2(vSize));
    console->MaxSize(Vector2(GetWidth()-10, GetHeight()-10));
}


void FractalInspectorApp::ZoomIn(float rate) {

    rate = rate * u_scale;

    u_scale_goal -= rate;
    // Stupid hack to make zoom originate at the center of the screen.
    u_translation_goal -= Vector2(0.5f, 0.5f)*rate;
}

void FractalInspectorApp::ZoomInTowards(const Vector2& dir, float rate) {
    rate = rate * u_scale;

    u_scale_goal -= rate;

    u_translation_goal -= Vector2(0.5f, 0.5f)*rate;

    u_translation_goal += dir*rate;
}

void FractalInspectorApp::ZoomOutTowards(const Vector2& dir, float rate) {
    rate = rate * u_scale;

    u_scale_goal += rate;
    // Stupid hack to make zoom originate at the center of the screen.
    u_translation_goal += Vector2(0.5f, 0.5f)*rate;

    u_translation_goal -= dir*rate;
}

void FractalInspectorApp::ZoomOut(float rate) {

    rate = rate * u_scale;

    u_scale_goal += rate;
    // Stupid hack to make zoom originate at the center of the screen.
    u_translation_goal += Vector2(0.5f, 0.5f)*rate;
}

void FractalInspectorApp::Update(float elapsed) {
    scene->Update(elapsed);

    if (IsKeyDown(Keys::F6)) {
        u_scale_goal = 2.f;
        u_translation_goal = {.4, 0};
    }

    // TODO: Vary the rate appropriately.
    float rate = elapsed * 0.5f;

    if (IsKeyDown(Keys::LeftShift))
        rate *= 0.125f;

    if (IsKeyDown(Keys::LeftControl))
        rate *= 0.125f;

    if (IsKeyDown(Keys::Minus))
        ZoomOut(rate);

    if (IsKeyDown(Keys::Equals))
        ZoomIn(rate);


    if (IsKeyDown(Keys::DownArrow))
        u_translation_goal.y += rate;
    if (IsKeyDown(Keys::UpArrow))
        u_translation_goal.y -= rate;

    if (IsKeyDown(Keys::LeftArrow))
        u_translation_goal.x += rate;
    if (IsKeyDown(Keys::RightArrow))
        u_translation_goal.x -= rate;


    u_time += elapsed;

    float lerp_rate = 0.1f;
    float julia_lerp_rate = 0.01f;

    u_translation = Vector2::Lerp(u_translation, u_translation_goal, lerp_rate);
    u_scale = Math::Lerp(u_scale, u_scale_goal, lerp_rate);

    u_julia_value = Vector2::Lerp(u_julia_value, u_julia_set, julia_lerp_rate);

    UpdateShaderUniforms(elapsed);

    int fps = 1.f / elapsed;


    // TODO: Consider moving this logic deeper into the shader to make it easier to work with.
    float readable_scale = Math::Round(ReadableScale(), 4);

    std::string fps_text = std::format("Pos: {},{} Zoom: {}x FPS: {}",
        Math::Round(u_translation.x, 4), Math::Round(u_translation.y, 4), readable_scale, fps);

    fps_label->Content(fps_text);



}

float FractalInspectorApp::ReadableScale() {
    return 3.f - u_scale;
}

void FractalInspectorApp::UpdateShaderUniforms(float elapsed) {
    Vector2 u_res = Vector2(GetSize().x, GetSize().y);

    // TODO: Why do we need to multiply X by 1.5 here?
    Vector2 u_mouse = Vector2(GetMouseCoordinates().x*1.5f, GetSize().y-GetMouseCoordinates().y);

    ;
    if (current_fractal == Fractal::JuliaSet) {
        current_shader = julia_shader;
        julia_shader->SetVector2("u_julia_set", u_julia_value);
    }

    if (current_fractal == Fractal::MandelbrotSet) {
       current_shader = mandelbrot_shader;
    }

    current_shader->SetVector2("u_resolution", u_res);
    current_shader->SetFloat("u_time", u_time);
    current_shader->SetVector2("u_mouse", u_mouse);
    current_shader->SetVector3("u_rgb_1", Vector3(u_rgb_1.RN(), u_rgb_1.GN(), u_rgb_1.BN()));
    current_shader->SetVector3("u_rgb_2", Vector3(u_rgb_2.RN(), u_rgb_2.GN(), u_rgb_2.BN()));
    current_shader->SetVector3("u_rgb_3", Vector3(u_rgb_3.RN(), u_rgb_3.GN(), u_rgb_3.BN()));
    current_shader->SetVector3("u_rgb_4", Vector3(u_rgb_4.RN(), u_rgb_4.GN(), u_rgb_4.BN()));
    current_shader->SetFloat("u_scale", u_scale);
    current_shader->SetVector2("u_translation", u_translation);
}

Vector2 solve(Vector2 a, Vector2 b) {
    float real = (a.x*a.x) + (-1 * a.y*a.y);
    float imag = (a.x*a.y)*2;

    real += b.x;
    imag += b.y;

    return {real, imag};
}

void FractalInspectorApp::Draw() {

    //Shader::SetDefault(); // TODO: Was this deprecated or replaced?

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // TODO: Have to be particular about the order-of-operations in regards to shaders.
    // Protip: Use RenderTarget to draw a quad with the shader in question. *After rendering the rest of the scene.*

    J2D::Begin();
    J2D::DrawRenderTarget(canvas, {0, 0});
    J2D::End();

    J2D::Begin();

    const float N = 512.f;
    const float B = 2.f;

    auto mpos_ipair = GetMouseCoordinates();
    Vector2 mpos (mpos_ipair.x, mpos_ipair.y);
    Vector2 R(GetWidth(), GetHeight());



    Vector2 trans = Vector2(0, R.y) - u_translation;

    Vector2 z = Vector2(0, 0);

    Vector2 c = ScreenSpaceToFractalSpace(mpos);
    Vector2 last_z = c;

    J2D::FillCircle(Colors::Red, FractalSpaceToScreenSpace(ScreenSpaceToFractalSpace(mpos)), 6);


    for (float i = 0.f; i < N; i++) {
        float x = z.x;
        float y = z.y;
        Vector2 z_nplus1 = Vector2(x*x - y*y, y*x + x*y) + c;


        J2D::DrawLine(Colors::White, FractalSpaceToScreenSpace(z), FractalSpaceToScreenSpace(last_z), 0.25);

        last_z = z;
        z = z_nplus1;




        if (Vector2::Dot(z, z) > B*B) break;
    }
    //
    J2D::End();

    scene->Draw();


    JGL::J2D::Begin(canvas, current_shader, true);

    JGL::J2D::FillRect(Colors::White, {0, 0}, Vector2(GetSize().x, GetSize().y));
    JGL::J2D::End();

}

void FractalInspectorApp::OnRefresh(float elapsed) {
    PropagateWindowSize();
    Update(elapsed);
    Draw();
    this->SwapBuffers();
}

enum JUI::MouseButton ToJUIEnum(const MouseButton& btn) {

    if (btn == MouseButtons::Left) return JUI::MouseButton::Left;
    if (btn == MouseButtons::Middle) return JUI::MouseButton::Middle;
    if (btn == MouseButtons::Right) return JUI::MouseButton::Right;

    // Default condition.
    return JUI::MouseButton::Left;
}

void FractalInspectorApp::OnMouseButtonDown(const ReWindow::MouseButtonDownEvent &e) {

    auto btn = ToJUIEnum(e.Button);

    if (scene->ObserveMouseInput(btn, true))
        return;

    if (btn == JUI::MouseButton::Right)
        panning = true;


    /*if (e.Button == MouseButtons::Left)
        scene->ObserveMouseInput(JUI::MouseButton::Left, true);
    if (e.Button == MouseButtons::Middle)
        scene->ObserveMouseInput(JUI::MouseButton::Middle, true);
    if (e.Button == MouseButtons::Right)
        scene->ObserveMouseInput(JUI::MouseButton::Right, true);*/
}

void FractalInspectorApp::OnMouseButtonUp(const ReWindow::MouseButtonUpEvent &e) {

    auto btn = ToJUIEnum(e.Button);

    if (scene->ObserveMouseInput(btn, false))
        return;

    if (btn == JUI::MouseButton::Right)
        panning = false;

    /*if (e.Button == MouseButtons::Left)
        scene->ObserveMouseInput(JUI::MouseButton::Left, false);
    if (e.Button == MouseButtons::Middle)
        scene->ObserveMouseInput(JUI::MouseButton::Middle, false);
    if (e.Button == MouseButtons::Right)
        scene->ObserveMouseInput(JUI::MouseButton::Right, false);*/
}

Vector2 last{0,0};

void FractalInspectorApp::OnMouseMove(const ReWindow::MouseMoveEvent &e) {
    Vector2 nmp = Vector2(e.Position.x, e.Position.y);
    if (scene->ObserveMouseMovement(nmp))
        return;

    // TODO: e.Delta is always 0, investigate in JUI.
    Vector2 delta = nmp - last;

    float base_slowdown = 768.f;
    base_slowdown = GetHeight(); // TODO: Compute the actual aspect ratio.

    // Invert vertical.
    Vector2 pan_direction(delta.x, -delta.y);
    float pan_rate = u_scale / base_slowdown;


    if (panning)
        u_translation_goal += pan_direction*pan_rate;

    last = nmp;
}

void FractalInspectorApp::OnKeyDown(const ReWindow::KeyDownEvent &e) {
    if (scene->ObserveKeyInput(e.key, true))
        return;

    if (e.key == Keys::R)
        ReloadShader();
}

void FractalInspectorApp::OnKeyUp(const ReWindow::KeyUpEvent &e) {
    scene->ObserveKeyInput(e.key, false);
}

void FractalInspectorApp::TakeScreenshot() {
    std::vector<Color4> pdatac4 = this->canvas->GetTexture()->GetPixelData();
    std::vector<Color3> pdatac3;
    pdatac3.reserve(pdatac4.size());
    for (auto Color:pdatac4) {
        pdatac3.emplace_back(Color.r, Color.g, Color.b);
    }
    uint8_t* frameBuffer = reinterpret_cast<uint8_t*>(pdatac3.data());

    time_t t = std::time(nullptr);
    tm tm = *std::localtime(&t);

    std::ostringstream filename;
    filename << std::put_time(&tm, "%Y-%m-%d %H-%M-%S") << ".bmp";

    std::ofstream file(filename.str(), std::ios::out | std::ios::binary);

    Vector2i wh = this->canvas->GetDimensions();

    int fileSize = 54 + 3 * wh.x * wh.y;
    unsigned char bmpfileheader[14] = {'B', 'M', 0, 0, 0, 0, 0, 0, 0, 0, 54, 0, 0, 0};
    unsigned char bmpinfoheader[40] = {40, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 24, 0};
    bmpfileheader[2] = (unsigned char) (fileSize);
    bmpfileheader[3] = (unsigned char) (fileSize >> 8);
    bmpfileheader[4] = (unsigned char) (fileSize >> 16);
    bmpfileheader[5] = (unsigned char) (fileSize >> 24);
    bmpinfoheader[4] = (unsigned char) (width);
    bmpinfoheader[5] = (unsigned char) (width >> 8);
    bmpinfoheader[6] = (unsigned char) (width >> 16);
    bmpinfoheader[7] = (unsigned char) (width >> 24);
    bmpinfoheader[8] = (unsigned char) (height);
    bmpinfoheader[9] = (unsigned char) (height >> 8);
    bmpinfoheader[10] = (unsigned char) (height >> 16);
    bmpinfoheader[11] = (unsigned char) (height >> 24);

    file.write((char*)bmpfileheader, 14);
    file.write((char*)bmpinfoheader, 40);

    int padding = (4 - (width * 3) % 4) % 4;
    for (int y = 0; y < height; ++y) {
        for (int x = 0; x < width; ++x) {
            unsigned char pixel[3];
            pixel[0] = frameBuffer[(y * width + x) * 3 + 2];
            pixel[1] = frameBuffer[(y * width + x) * 3 + 1];
            pixel[2] = frameBuffer[(y * width + x) * 3];
            file.write((char*)pixel, 3);
        }
        for (int p = 0; p < padding; ++p) {
            file.write((char*)"\0", 1);
        }
    }

    file.close();

    console->Log(std::format("Took screenshot: {}", filename.str()));

}
